@@ -1,4 +1,4 @@
-# Copyright (c) 2018-2020 Carnegie Mellon University
+# Copyright (c) 2018-2021 Carnegie Mellon University
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy of this
 # software and associated documentation files (the "Software"), to deal in the Software
@@ -37,7 +37,10 @@ may be strictly algorithmic, may interact with human subjects, or may be embedde
 sites.
 """
 
-__version__ = "1.0.9"
+__version__ = "1.1.1"
+
+if "dev" in __version__:
+    print("PyACTUp version", __version__)
 
 import collections
 import collections.abc as abc
@@ -49,6 +52,7 @@ import random
 import sys
 
 from collections import OrderedDict
+from contextlib import contextmanager
 from warnings import warn
 
 __all__ = ("Memory", "set_similarity_function", "use_actr_similarity")
@@ -98,8 +102,11 @@ class Memory(dict):
                  retrieval_time_increment=DEFAULT_RETRIEVAL_TIME_INCREMENT,
                  optimized_learning=False):
         self._temperature_param = 1 # will be reset below, but is needed for noise assignment
-        self.noise = noise
+        self._activation_noise_cache = None
+        self._activation_noise_cache_time = None
+        self._noise = None
         self._decay = None
+        self.noise = noise
         self._optimized_learning = False
         self.decay = decay
         if temperature is None and not self._validate_temperature(None, noise):
@@ -115,7 +122,7 @@ class Memory(dict):
         self.reset(optimized_learning=bool(optimized_learning))
 
     def __repr__(self):
-        return f"<Memory {dict(self.values())}>"
+        return f"<Memory {dict(self.items())}>"
 
     def __str__(self):
         return f"<Memory {id(self)}>"
@@ -143,6 +150,78 @@ class Memory(dict):
             for k, v in preserved.items():
                 v._references = 0 if self._optimized_learning else [0]
                 self[k] = v
+        self._clear_noise_cache()
+
+    def _clear_noise_cache(self):
+        if self._activation_noise_cache is not None:
+            self._activation_noise_cache.clear()
+            self._activation_noise_cache_time = self._time
+
+    @property
+    @contextmanager
+    def fixed_noise(self):
+        """A context manager used to force multiple activations of a given chunk at the
+        same time to use the same activation noise.
+
+        .. warning::
+            Use of ``fixed_noise`` is rarely appropriate, and easily leads to biologically
+            implausible results. It is provided only for esoteric purposes. When its use
+            is required it should be wrapped around the smallest fragment of code
+            practical.
+
+        >>> m = Memory()
+        >>> m.learn(color="red")
+        True
+        >>> m.activation_history = []
+        >>> m.retrieve()
+        <Chunk 0000 {'color': 'red'}>
+        >>> m.retrieve()
+        <Chunk 0000 {'color': 'red'}>
+        >>> pprint(m.activation_history)
+        [OrderedDict([('name', '0000'),
+                      ('creation_time', 0),
+                      ('attributes', (('color', 'red'),)),
+                      ('references', (0,)),
+                      ('base_activation', 0.0),
+                      ('activation_noise', -0.5560365037582069),
+                      ('activation', -0.5560365037582069)]),
+         OrderedDict([('name', '0000'),
+                      ('creation_time', 0),
+                      ('attributes', (('color', 'red'),)),
+                      ('references', (0,)),
+                      ('base_activation', 0.0),
+                      ('activation_noise', 0.3352040096847542),
+                      ('activation', 0.3352040096847542)])]
+        >>> m.activation_history = []
+        >>> with m.fixed_noise:
+        ...     m.retrieve()
+        ...     m.retrieve()
+        ...
+        <Chunk 0000 {'color': 'red'}>
+        <Chunk 0000 {'color': 'red'}>
+        >>> pprint(m.activation_history)
+        [OrderedDict([('name', '0000'),
+                      ('creation_time', 0),
+                      ('attributes', (('color', 'red'),)),
+                      ('references', (0,)),
+                      ('base_activation', 0.0),
+                      ('activation_noise', 0.9791851742870211),
+                      ('activation', 0.9791851742870211)]),
+         OrderedDict([('name', '0000'),
+                      ('creation_time', 0),
+                      ('attributes', (('color', 'red'),)),
+                      ('references', (0,)),
+                      ('base_activation', 0.0),
+                      ('activation_noise', 0.9791851742870211),
+                      ('activation', 0.9791851742870211)])]
+        """
+        self._activation_noise_cache = {}
+        self._activation_noise_cache_time = self._time
+        try:
+            yield self
+        finally:
+            self._activation_noise_cache = None
+            self._activation_noise_cache_time = None
 
     @property
     def time(self):
@@ -158,9 +237,50 @@ class Memory(dict):
         """
         if amount < 0:
             raise ValueError(f"Time cannot be advanced backward ({amount})")
-        self._time += amount
+        if amount:
+            self._time += amount
+            self._clear_noise_cache()
         return self._time
 
+    @property
+    @contextmanager
+    def current_time(self):
+        """A context manager used to allow reverting to the current time after advancing
+        it and simiulating retrievals or similar operations in the future.
+
+        .. warning::
+            It is rarely appropriate to use ``current_time``. When it is used, care should
+            be taken to avoid creating biologically implausible models. Also, learning
+            within a ``current_time`` context will typically lead to tears as having
+            chunks created or reinforced in the future results in failures of attempts to
+            retrieve them.
+
+        >>> m = Memory(temperature=1, noise=0)
+        >>> m.learn(size=1)
+        True
+        >>> m.advance(10)
+        11
+        >>> m.learn(size=10)
+        True
+        >>> m.blend("size")
+        7.983916860341838
+        >>> with m.current_time as t:
+        ...     m.advance(10_000)
+        ...     m.blend("size")
+        ...     (t, m.time)
+        ...
+        10012
+        5.501236696240907
+        (12, 10012)
+        >>> m.time
+        12
+        """
+        old = self._time
+        try:
+            yield old
+        finally:
+            self._time = old
+
     @property
     def learning_time_increment(self):
         """The default amount of time to :meth:`advance` by after performing a learn
@@ -214,7 +334,9 @@ class Memory(dict):
                 self.temperature = 1
             else:
                 self._temperature = t
-        self._noise = value
+        if value != self._noise:
+            self._noise = value
+            self._clear_noise_cache()
 
     @property
     def decay(self):
@@ -317,6 +439,11 @@ class Memory(dict):
         are called as necessary, and
         the resulting values are multiplied by the mismatch penalty and subtracted
         from the activation.
+
+        Attributes for which no similarity function has been defined are always compared
+        exactly, and chunks not matching on this attributes are not included at all in the
+        corresponding partial retrievals or blending operations.
+
         Attempting to set this parameter to a value other than ``None`` or a real number
         raises a :exc:`ValueError`.
         """
@@ -406,12 +533,17 @@ class Memory(dict):
     _similarity_functions = {}
     _similarity_cache = pylru.lrucache(SIMILARITY_CACHE_SIZE)
 
+    # possibly update docstrings for retrieve() and blend() to reflect new similarity stuff,
+    # or maybe put it all n set_similarity_function() and/or mismatch?
     def _similarity(self, x, y, attribute):
         # always returns the "natural" similarity
+        # returns None if similarity is inapplicatble to attribute
         if x == y:
             return 1
         fn = self._similarity_functions.get(attribute)
-        if fn:
+        if fn is True:
+            return 0
+        elif fn:
             signature = (x, y, attribute)
             result = self._similarity_cache.get(signature)
             if result is not None:
@@ -430,8 +562,6 @@ class Memory(dict):
                 result += 1
             self._similarity_cache[signature] = result
             return result
-        else:
-            return 0
 
     def learn(self, advance=None, **kwargs):
         """Adds, or reinforces, a chunk in this Memory with the attributes specified by *kwargs*.
@@ -557,8 +687,8 @@ class Memory(dict):
             return result
         finally:
             if old is not None:
-                # don't advance if there's an error or for some other reason we don't
-                # finish normally
+                # Don't advance if there's an error or for some other reason we don't
+                # finish normally; note that the noise cache is still cleared, though.
                 self._time = old
 
     def _exact_match(self, conditions):
@@ -580,11 +710,19 @@ class Memory(dict):
                     best_activation = a
         return best_chunk
 
-    def _make_noise(self):
+    def _make_noise(self, chunk):
         if not self._noise:
             return 0
-        p = random.uniform(sys.float_info.epsilon, 1 - sys.float_info.epsilon)
-        return self._noise * math.log((1.0 - p) / p)
+        if self._activation_noise_cache is not None:
+            result = self._activation_noise_cache.get(chunk._name)
+        else:
+            result = None
+        if result is None:
+            p = random.uniform(sys.float_info.epsilon, 1 - sys.float_info.epsilon)
+            result = self._noise * math.log((1.0 - p) / p)
+        if self._activation_noise_cache is not None:
+            self._activation_noise_cache[chunk._name] = result
+        return result
 
     class _Activations(abc.Iterable):
 
@@ -600,27 +738,37 @@ class Memory(dict):
             while True:
                 chunk = self._chunks.__next__()             # pass on up the Stop Iteration
                 if self._conditions.keys() <= chunk.keys(): # subset
-                    if self._memory._mismatch is not None:
-                        activation = chunk._activation(True)
-                        mismatch = (self._memory._mismatch
-                                    * sum(self._memory._similarity(c, chunk[s], s) - 1
-                                          for s, c in self._conditions.items()))
-                        total = activation + mismatch
-                        if self._memory._activation_history is not None:
-                            history = self._memory._activation_history[-1]
-                            history["mismatch"] = mismatch
-                            history["activation"] = total
-                        return (chunk, total)
+                    if self._memory._mismatch is None:
+                        exact = self._conditions.keys()
+                        partial = []
                     else:
-                        if not all(chunk[a] == v for a, v in self._conditions.items()):
-                            continue
-                        activation = chunk._activation(True)
+                        exact = []
+                        partial = []
+                        for c in self._conditions.keys():
+                            if Memory._similarity_functions.get(c):
+                                partial.append(c)
+                            else:
+                                exact.append(c)
+                    if not all(chunk[a] == self._conditions[a] for a in exact):
+                        continue
+                    activation = chunk._activation(True)
+                    if self._memory._mismatch is None:
                         if self._memory._activation_history is not None:
                             self._memory._activation_history[-1]["activation"] = activation
                         return (chunk, activation)
+                    mismatch = (self._memory._mismatch
+                                * sum(self._memory._similarity(self._conditions[a], chunk[a], a) - 1
+                                      for a in partial))
+                    total = activation + mismatch
+                    if self._memory._activation_history is not None:
+                        history = self._memory._activation_history[-1]
+                        history["mismatch"] = mismatch
+                        history["activation"] = total
+                    return (chunk, total)
+
 
     def _activations(self, conditions):
-        return self._Activations(self, conditions)
+         return self._Activations(self, conditions)
 
     def _partial_match(self, conditions):
         best_chunks = []
@@ -683,15 +831,15 @@ class Memory(dict):
                 return None
         finally:
             if old is not None:
-                # don't advance if there's an error or for some other reason we don't
-                # finish normally
+                # Don't advance if there's an error or for some other reason we don't
+                # finish normally; note that the noise cache is still cleared, though.
                 self._time = old
 
     def best_blend(self, outcome_attribute, iterable, select_attribute=None, advance=None, minimize=False):
         """Returns two values (as a 2-tuple), describing the extreme blended value of the *outcome_attribute* over the values provided by *iterable*.
         The extreme value is normally the maximum, but can be made the minimum by setting
         *minimize* to True. The values returned by *iterable* should be dictionary-like
-        objects that can be passed as the *kwargs* argument to :meth:`blend`. The first
+        object that can be passed as the *kwargs* argument to :meth:`blend`. The first
         return value is the *kwargs* value producing the best blended value, and second is
         that blended value. If there is a tie, with two or more *kwargs* values all
         producing the same, best blended value, then one of them is chosen randomly. If
@@ -759,7 +907,7 @@ class Memory(dict):
         finally:
             if old is not None:
                 # Don't advance if there's an error or for some other reason we don't
-                # finish normally
+                # finish normally; note that the noise cache is still cleared, though.
                 self._time = old
 
 
@@ -797,6 +945,9 @@ def set_similarity_function(function, *slots):
     No error is raised if either of these constraints is violated, but the results
     will, in most cases, be meaningless if they are.
 
+    If ``True`` is supplied as the *function* a default similarity function is used that
+    returns one if its two arguments are ``==`` and zero otherwise.
+
     >>> def f(x, y):
     ...     if y < x:
     ...         return f(y, x)
@@ -804,7 +955,12 @@ def set_similarity_function(function, *slots):
     >>> set_similarity_function(f, "length", "width")
     """
     for s in slots:
-        Memory._similarity_functions[s] = function
+        if callable(function):
+            Memory._similarity_functions[s] = function
+        elif function:
+            Memory._similarity_functions[s] = True
+        elif s in Memory._similarity_functions:
+            del Memory._similarity_functions[s]
     Memory._similarity_cache.clear()
 
 
@@ -834,7 +990,7 @@ class Chunk(dict):
     def _activation(self, for_partial=False):
         # Does not include the mismatch penalty component, that's handled by the caller.
         base = self._get_base_activation() if self._memory._decay is not None else 0
-        noise = self._memory._make_noise()
+        noise = self._memory._make_noise(self)
         result = base + noise
         if self._memory._activation_history is not None:
             history = OrderedDict(name=self._name,
@@ -851,7 +1007,7 @@ class Chunk(dict):
         return result
 
     # Note that memoizing expt and ln doesn't make much difference, but it does speed
-    # things up a tiny bit, most noticeably under PyPy
+    # things up a tiny bit, most noticeably under PyPy.
     def _cached_expt(self, base):
         try:
             result = self._memory._expt_cache[base]
